/////////////////////////////////////////////////////////////////////////////////
//                  CONFIDENTIAL AND PROPRIETARY INFORMATION
//             (C) 2003, 2004 Synopsys Inc (Unpublished) and Chris Spear
//                             All Rights Reserved
//
// This file can be freely distributed providing this (C) notice remains
/////////////////////////////////////////////////////////////////////////////////

/* real.c - Library of real functions used by Vera through DirectC

   v1.9  9/01/04  Support NTB
   v1.8 11/17/03  Fix mod bug in R2S
   v1.7  7/24/03  Add multiple buffers to R2S
   v1.6  7/23/03  Radd takes 10 args, added R0, R1, Rm1
   v1.5  3/14/03  Pack the REAL into the Vera's c/d bits
   v1.4  3/05/03  Speed up Req with pow10, added bounds check
   v1.3  2/25/03  Added Rfrac, Reqd, parameterized real type
   v1.2  2/19/03  Added digits to Req, Rne, formatted R2S
   v1.1  2/11/03  Added Rdelta, Rpi, Re, Rhypot
   v1.0  2/ 7/03  Created

   In C I am using a double which is 64 bits long
   In Vera I am using a reg[31:0].  This is stored in a vec32
   array, where the data is in vec32.d and vec32.c


   Limitations:
   - The R2S function has only one buffer.  Multiple calls in a 
   printf seem to work, but could cause problems in string concat.
   - What is the maximum width of a printed real? 512 chars?

   Wishlist:
   - Allow switch between double and long double by changing the
   definition of REAL here and real.vri  Several attempts have failed...


 */

#ifdef SYNOPSYS_NTB
#include "vc_hdrs.h"
#else
#include "tb.dc.h"
#endif
// In case you need the Vera DirectC data structures
//#include <vera_directc.h>


#include <math.h>

// If you change the following, change the REAL definition in real.vri
// Does not work yet...
#define REAL double
//#define REAL long double

#define REAL_C_SIZE    sizeof(REAL) / sizeof(int)

// Size of a real in Vera
#define REAL_VERA_SIZE REAL_C_SIZE

/*The real heap is used to store multiple REAL values
  This way there can be N operations active at once, without
  stepping on the temporaries.
  A Vera REAL is 64 bits, stored as c/d/c/d, 4 int's
*/

#define HEAP_SIZE   64
#define HEAP_LENGTH HEAP_SIZE * REAL_VERA_SIZE
static int REAL_heap[HEAP_LENGTH] = {0};
static int REAL_heap_idx = 0;
// This macro returns the next location in the heap
#define HEAP_NEXT_REAL(p) { \
   p = (REAL *) &REAL_heap[REAL_heap_idx]; \
   REAL_heap_idx = (REAL_heap_idx+1) % HEAP_LENGTH; }

// Macro to copy a C REAL into a Vera REAL (reg[63:0])
#define R2V(v,r) {HEAP_NEXT_REAL(v); \
		  *v = (r);}


// NTB: Return a value by assigning it to the result argument
#define RETURNN(dv) \
{vec32 *dd = (vec32 *) &dv; \
 result->c = dd->c; \
 result->d = dd->d; }

// Vera: Return a value by putting in a temporary, and returning it's address
#define RETURNV(dv) \
{REAL *dresult; \
 R2V(dresult, (dv)); \
 return (vec32 *) dresult; }


// Speed up Req by eliminating the call: pow(10, 0.0-digit)
REAL pow10[30] = {1e-00, 1e-01, 1e-02, 1e-03, 1e-04, 1e-05, 1e-06, 1e-07, 1e-08, 1e-09,
		  1e-10, 1e-11, 1e-12, 1e-13, 1e-14, 1e-15, 1e-16, 1e-17, 1e-18, 1e-19,
		  1e-20, 1e-21, 1e-22, 1e-23, 1e-24, 1e-25, 1e-26, 1e-27, 1e-28, 1e-29};


////////////////////////////////////////////////////////////////////////

// Convert an integer to a Vera REAL
vec32 *I2R(int i) {
  REAL d = i;
  RETURNV(d);
  // printf("C: I2R_c(%d) dresult=%f @ %x, d=%f\n", i, *dresult, dresult, d);
}

// Task to convert an integer into a Vera REAL
void I2R_c(vec32 *result, int i) {
  REAL d = i;
  RETURNN(d);
  // printf("C: I2R_c(%0d) d=%f result=%x,%x\n", i, d, result->c, result->d);
}


// Convert a Vera REAL to integer
int R2I(vec32 *va) {
  REAL *ra = (REAL *) va;
  int i = *ra;
  return i;
}


// Convert a Vera REAL into a string
// You could get into trouble calling this more than once in
// a single printf, but it seems to work
#define MAX_STR 128
#define MAX_ARGS 16
char *R2S(vec32 *vsrc, char *fmt) {
  REAL *src = (REAL *) vsrc;
  static char dst[MAX_ARGS][MAX_STR];
  char *p;
  static int didx = 0;
  REAL d;

  if (fmt[0] == 0) // No format given
    fmt = "%f";
  snprintf(dst[didx], MAX_STR, fmt, *src);
  p = dst[didx];
  didx = (didx+1) % MAX_ARGS;
  return p;
}


// Add several Vera REALs and return a Vera REAL
// Really should use an array for arguments
vec32 *Radd(vec32 *va0, vec32 *va1, vec32 *va2, vec32 *va3, vec32 *va4, vec32 *va5, vec32 *va6, vec32 *va7, vec32 *va8, vec32 *va9 ) {
  REAL d, *ra0 = (REAL *) va0, *ra1 = (REAL *) va1;

  d = *ra0 + *ra1;
  // In Vera arguments are always defined - but just to be safe...
  if (va2) {REAL *ra2 = (REAL *) va2; d += *ra2; }
  if (va3) {REAL *ra3 = (REAL *) va3; d += *ra3; }
  if (va4) {REAL *ra4 = (REAL *) va4; d += *ra4; }
  if (va5) {REAL *ra5 = (REAL *) va5; d += *ra5; }
  if (va6) {REAL *ra6 = (REAL *) va6; d += *ra6; }
  if (va7) {REAL *ra7 = (REAL *) va7; d += *ra7; }
  if (va8) {REAL *ra8 = (REAL *) va8; d += *ra8; }
  if (va9) {REAL *ra9 = (REAL *) va9; d += *ra9; }
  RETURNV(d);
}


// Add several Vera REALs and return a Vera REAL
// Really should use an array for arguments
void Radd_c(vec32 *result, vec32 *va0, vec32 *va1, vec32 *va2, vec32 *va3, vec32 *va4, vec32 *va5, vec32 *va6, vec32 *va7, vec32 *va8, vec32 *va9 ) {
  REAL dr, *ra0 = (REAL *) va0, *ra1 = (REAL *) va1;

  dr = *ra0 + *ra1;
  // In Vera arguments are always defined - but just to be safe...
  if (va2) {REAL *ra2 = (REAL *) va2; dr += *ra2; }
  if (va3) {REAL *ra3 = (REAL *) va3; dr += *ra3; }
  if (va4) {REAL *ra4 = (REAL *) va4; dr += *ra4; }
  if (va5) {REAL *ra5 = (REAL *) va5; dr += *ra5; }
  if (va6) {REAL *ra6 = (REAL *) va6; dr += *ra6; }
  if (va7) {REAL *ra7 = (REAL *) va7; dr += *ra7; }
  if (va8) {REAL *ra8 = (REAL *) va8; dr += *ra8; }
  if (va9) {REAL *ra9 = (REAL *) va9; dr += *ra9; }
  RETURNN(dr);
}


// Divide two Vera REALs and return a Vera REAL
vec32 *Rdiv(vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = *ra / *rb;
  RETURNV(d);
}

void Rdiv_c(vec32 *result, vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb  = (REAL *) vb, 
    dr = *ra / *rb;
  RETURNN(dr);
}


// Compute the modulo of two Vera REALs and return a Vera REAL
// What does this really do?
vec32 *Rmod(vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb,
    d = fmod(*ra, *rb);
  RETURNV(d);
}

void Rmod_c(vec32 *result, vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = fmod(*ra, *rb);
  RETURNN(d);
}


// Multiply two Vera REALs and return a Vera REAL
vec32 *Rmul(vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = *ra * *rb;
  RETURNV(d);
}

void Rmul_c(vec32 *result, vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = *ra * *rb;
  RETURNN(d);
}



// Subtract two Vera REALs and return a Vera REAL
vec32 *Rsub(vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = *ra - *rb;
  RETURNV(d);
}

void Rsub_c(vec32 *result, vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = *ra - *rb;
  RETURNN(d);
}


// Delta: Subtract two Vera REALs, return the absolute value of the Vera REAL
vec32 *Rdelta(vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = fabs(*ra - *rb);
  RETURNV(d);
}

void Rdelta_c(vec32 *result, vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = fabs(*ra - *rb);
  RETURNN(d);
}



// Return the square of the Vera REAL
vec32 *Rsquare(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = *ra * *ra;
  RETURNV(d);
}

void Rsquare_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = *ra * *ra;
  RETURNN(d);
}



// Return the cube of the Vera REAL
vec32 *Rcube(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = *ra * *ra * *ra;
  RETURNV(d);
}

void Rcube_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = *ra * *ra * *ra;
  RETURNN(d);
}


// Divide two Vera integers and return a Vera REAL
// Sort of an extra-funtion I2R
vec32 *Rfract(int a, int b) {
  REAL ra = a, rb = b,
    d = ra / rb;
  RETURNV(d);
}

void Rfract_c(vec32 *result, int a, int b) {
  REAL ra = a, rb = b, 
    d = ra / rb;
  RETURNN(d);
}


// Compute the arccos of a Vera REAL and return a Vera REAL
vec32 *Racos(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = acos(*ra);
  RETURNV(d);
}

void Racos_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = acos(*ra);
  RETURNN(d);
}


// Compute the arcsin of a Vera REAL and return a Vera REAL
vec32 *Rasin(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = asin(*ra);
  RETURNV(d);
}

void Rasin_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = asin(*ra);
  RETURNN(d);
}



// Compute the arctan of a Vera REAL and return a Vera REAL
vec32 *Ratan(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = atan(*ra);
  RETURNV(d);
}

void Ratan_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = atan(*ra);
  RETURNN(d);
}



// Compute the cosine of a Vera REAL and return a Vera REAL
vec32 *Rcos(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = cos(*ra);
  RETURNV(d);
}

void Rcos_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = cos(*ra);
  RETURNN(d);
}



// Compute the sine of a Vera REAL and return a Vera REAL
vec32 *Rsin(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = sin(*ra);
  RETURNV(d);
}

void Rsin_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = sin(*ra);
  RETURNN(d);
}



// Compute the tan of a Vera REAL and return a Vera REAL
vec32 *Rtan(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = tan(*ra);
  RETURNV(d);
}

void Rtan_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = tan(*ra);
  RETURNN(d);
}



// REAL = sqrt(a**2 + b**2)   Compute the hypotenuse of a right triangle
// From Janick Bergeron
vec32 *Rhypot(vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = sqrt(*ra**ra + *rb**rb);
  RETURNV(d);
}

void Rhypot_c(vec32 *result, vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = sqrt(*ra**ra + *rb**rb);
  RETURNN(d);
}


// Compute the log of a Vera REAL and return a Vera REAL
vec32 *Rlog(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = log(*ra);
  RETURNV(d);
}

void Rlog_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = log(*ra);
  RETURNN(d);
}


// Compute the log10 of a Vera REAL and return a Vera REAL
vec32 *Rlog10(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = log10(*ra);
  RETURNV(d);
}

void Rlog10_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = log10(*ra);
  RETURNN(d);
}



// Raise a Vera REAL to the power of another and return a Vera REAL
vec32 *Rpow(vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = pow(*ra, *rb);
  RETURNV(d);
}

void Rpow_c(vec32 *result, vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb, 
    d = pow(*ra, *rb);
  RETURNN(d);
}



// Compute the squareroot of a Vera REAL and return a Vera REAL
vec32 *Rsqrt(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = sqrt(*ra);
  RETURNV(d);
}

void Rsqrt_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = sqrt(*ra);
  RETURNN(d);
}



// Return the absolute value of a Vera REAL
vec32 *Rabs(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = fabs(*ra);
  RETURNV(d);
}

void Rabs_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = fabs(*ra);
  RETURNN(d);
}



// Return the ceil of a Vera REAL
vec32 *Rceil(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = ceil(*ra);
  RETURNV(d);
}

void Rceil_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = ceil(*ra);
  RETURNN(d);
}


// Return the floor of a Vera REAL
vec32 *Rfloor(vec32 *va) {
  REAL *ra = (REAL *) va,
    d = floor(*ra);
  RETURNV(d);
}

void Rfloor_c(vec32 *result, vec32 *va) {
  REAL *ra = (REAL *) va, 
    d = floor(*ra);
  RETURNN(d);
}



// Compare two Vera REALs within a delta and return an integer = a ~!= b
// if digits==0 (default) do a straight compare
// Req(a, b, digits) = (a - b) > a * (10 ** -digits)
int Req(vec32 *va, vec32 *vb, int digits) {
  REAL dd, *ra = (REAL *) va, *rb = (REAL *) vb;

  if (digits<0 || digits>29) {
    printf("*ERROR* Req (real.c) digits=%d, must be >0 and <30\n", digits);
    return 0;
  }

  if (digits == 0)
      return *ra == *rb;
  else {
    // Scale "a" by the number of digits to compare
    dd = pow10[digits] * *ra;
    //printf("C fabs(%f-%f) %f ? dd %f\n", *ra, *b, fabs(*ra-*b), dd);
    return fabs(*ra-*rb) < dd;
  }
} // Req


// Compare two Vera REALs within a delta and return an integer telling
// how many places matched
// digits = Reqd(a, b) = (a - b) < a * (10 ** -digits)
int Reqd(vec32 *va, vec32 *vb) {
  REAL dd, *ra = (REAL *) va, *rb = (REAL *) vb;
  REAL dfabs, ddst;
  int digits = 0;

  dfabs = fabs(*ra-*rb);
  dd = *ra / 10.0;
  while (digits<30 && dfabs < dd) {
    digits++;
    dd /= 10.0;
  }
  return digits;
} // Reqd


// Compare two Vera REALs within a delta and return an integer = a ~!= b
// if digits==0 (default) do a straight compare
// Rne(a, b, digits) = (a - b) > a * (10 ** -digits)
int Rne(vec32 *va, vec32 *vb, int digits) {
  REAL dd, *ra = (REAL *) va, *rb = (REAL *) vb;

  if (digits<0 || digits>29) {
    printf("*ERROR* Rne (real.c) digits=%d, must be >0 and <30\n", digits);
    return 0;
  }


  if (digits == 0)
      return *ra == *rb;
  else {
    dd = pow10[digits];
    return fabs(*ra-*rb) > (dd**ra);
  }
}


// Compare two Vera REALs and return an integer = a >= b
int Rge(vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb;
  REAL ddst;

  return *ra >= *rb;
}


// Compare two Vera REALs and return an integer = a > b
int Rgt(vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb;
  REAL ddst;

  return *ra > *rb;
}


// Compare two Vera REALs and return an integer = a <= b
int Rle(vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb;
  REAL ddst;

  return *ra <= *rb;
}


// Compare two Vera REALs and return an integer = a > b
int Rlt(vec32 *va, vec32 *vb) {
  REAL *ra = (REAL *) va, *rb = (REAL *) vb;
  REAL ddst;
  return *ra < *rb;
}


// Return pi as a Vera REAL
vec32 *Rpi() {
  static REAL pi = M_PI;
  return (vec32 *) &pi;
}

void Rpi_c(vec32 *result) {
  REAL d = M_PI;
  RETURNN(d);
}


// Return e as a Vera REAL
vec32 *Re() {
  static REAL e = M_E;
  return (vec32 *) &e;
}

void Re_c(vec32 *result) {
  REAL d = M_E;
  RETURNN(d);
}


// Return 0 as a Vera REAL
vec32 *R0() {
  static REAL zero = 0.0;
  return (vec32 *) &zero;
}

void R0_c(vec32 *result) {
  REAL d = 0.0;
  RETURNN(d);
}



// Return 1.0 as a Vera REAL
vec32 *R1() {
  static REAL one = 1.0;
  return (vec32 *) &one;
}

void R1_c(vec32 *result) {
  REAL d = 1.0;
  RETURNN(d);
}


// Return -1.0 as a Vera REAL
vec32 *Rm1() {
  static REAL mone = -1.0;
  return (vec32 *) &mone;
}

void Rm1_c(vec32 *result) {
  REAL d = -1.0;
  RETURNN(d);
}


// Convert a string into a Vera REAL
vec32 *S2R(char *src) {
  REAL *dresult;
  REAL d;

  sscanf(src, "%le", &d);
  // printf("C: S2R '%s' scanned to %f\n", src, d);
  R2V(dresult, d);
  return (vec32 *) dresult;
}



// Convert a string into a Vera REAL
void  S2R_c(vec32 *result, char *src) {
  REAL d;

  sscanf(src, "%le", &d);
  RETURNN(d);
  // printf("C: S2R_c '%s' scanned to %f\n", src, d);
}
